from django.contrib.staticfiles.finders import get_finders
from django.core.management.base import BaseCommand

import sass
import time
import os


class Command(BaseCommand):
    """
    Using libsass, compile all .scss files found in the paths defined in settings.STATICFILES_FINDERS.
    """
    help = "Using libsass, compile all .scss files found in the paths defined in settings.STATICFILES_FINDERS."

    @staticmethod
    def add_arguments(parser):

        parser.add_argument(
            '-s',
            '--style',
            dest='style',
            type=str,
            choices=['expanded', 'nested', 'compact', 'compressed'],
            default='expanded',
            help=('An optional coding style of the compiled result. '
                  'Choose one of: `\'nested\'`, `\'expanded\'` (default), `\'compact\'`, `\'compressed\'`'),
        )

        parser.add_argument(
            '-p',
            '--precision',
            dest='precision',
            type=int,
            default=8,
            help='Optional precision for numbers. 8 by default.',
        )

        parser.add_argument(
            '-nb',
            '--no-build',
            dest='build',
            action='store_true',
            default=False,
            help=('Don\'t create build folder, e.g. "app/static/app/css/style.css" '
                  'instead "app/static/app/build/css/style.css".'),
        )

        parser.add_argument(
            '-m',
            '--map',
            dest='map',
            action='store_true',
            default=False,
            help='Build a source map "style.css.map".',
        )

        parser.add_argument(
            '-c',
            '--clean',
            dest='clean',
            action='store_true',
            default=False,
            help=('Remove "style.css", "style.css.map", "style.min.css", "style.min.css.map" files before '
                  'compilation. This action will only take effect on current destination folder.'),
        )

        parser.add_argument(
            '-w',
            '--watch',
            dest='watch',
            action='store_true',
            default=False,
            help='Watch and compile files when scss files are changed.',
        )

    def handle(self, *args, **options):
        """
        Runs Sass Compiler once time or in watch mode.
        :param args:
        :param options:
        """
        style = options.get('style', 'expanded')            # expanded | nested | compact | compressed
        precision = options.get('precision', 8)             # 8
        build = not options.get('build', False)             # True | False
        map = options.get('map', False)                     # True | False
        clean = options.get('clean', False)                 # True | False
        include_paths = self.get_static_paths()             # List of paths to find @import

        if options.get('watch', False):
            self.stdout.write('Start watching...')
            try:
                files = {}
                while True:
                    files, recompile = self.needs_recompile(files)
                    if recompile:
                        self.compile_files(style, build, include_paths, precision, map, clean)
                    time.sleep(1)

            except (InterruptedError, KeyboardInterrupt):
                self.stdout.write('Stop watching.')

        else:
            self.compile_files(style, build, include_paths, precision, map, clean)

        self.stdout.write('End.')

    def compile_files(self, output_style, build, include_paths, precision, source_map, clean):
        """
        Finds all scss files in all apps (except who starts with _ e.g.: _variable.scss) and runs sass compile
        :param output_style:        'expanded' | 'nested' | 'compact' | 'compressed'
        :param build:               True | False
        :param include_paths:       List of paths to find @import
        :param precision:           8
        :param source_map:          True | False
        :param clean:               True | False
        """
        for scss_filename in self.get_scss_files(['_*.scss']):

            css_filename, source_map_filename = self.get_output_files(scss_filename, build, output_style)

            try:
                output = sass.compile(
                    filename=scss_filename,
                    source_map_filename=source_map_filename,
                    include_paths=include_paths,
                    output_style=output_style,
                    precision=precision,
                    omit_source_map_url=not source_map
                )

                if output and isinstance(output, tuple):
                    self.clean_files(css_filename) if clean else None
                    self.store_file(css_filename, output[0])
                    self.store_file(source_map_filename, output[1]) if source_map and len(output) > 1 else None

            except sass.CompileError as exc:
                self.stdout.write(str(exc))

        self.stdout.write("Files compiled at %s" % time.time())

    def needs_recompile(self, files):
        """
        Check if any files have been changed and compilation is necessary
        Return dict with file and modified datetime
        :param files: {'Your/Path/project/apps/app1/static/app1/scss/style.scss: 1587830257.0483491, ...}
        :return:
        True|False,
        {
            'Your/Path/project/apps/app1/static/app1/scss/style.scss: 1587830257.0483491
            ...
        }
        """
        update = False
        for file in self.get_scss_files():
            modified = os.path.getmtime(file)
            if modified > files.get(file, 0):
                files.update({file: modified})
                update = True

        return files, update

    @staticmethod
    def get_output_files(filename, build, style):
        """
        Gets output css and source map file names
        Manage the output path according to the arguments --no-build and --style
        :param filename:    'Your/Path/project/apps/app1/static/app1/scss/style.scss'
        :param build:       True | False
        :param style:       'expanded' | 'nested' | 'compact' | 'compressed'
        :return:
        'Your/Path/project/apps/app1/static/app1/build/css/style.css',
        'Your/Path/project/apps/app1/static/app1/build/css/style.css.map'
        """
        scss_folder = ''.join([os.sep, 'scss', os.sep])                                 # \scss\
        build_folder = ''.join([os.sep, 'build']) if build else ''                      # \build | ''
        css_folder = ''.join([build_folder, os.sep, 'css', os.sep])                     # \build\css\ | \css\
        css_file = filename.replace(scss_folder, css_folder).replace('.scss', '.css')   # ...\build\css\style.css
        source_map_filename = ''.join([css_file, '.map'])                               # ...\build\css\style.css.map
        if style == 'compressed':
            css_file = css_file.replace('.css', '.min.css')                             # ...\build\css\style.min.css

        return css_file, source_map_filename

    @staticmethod
    def get_scss_files(ignore_patterns=[]):
        """
        Get scss files
        :param ignore_patterns: List ['_*.scss'] Exclude all files who starts with _ and end .scss e.g: _variables.scss
        :return:
        [
            'Your/Path/project/apps/app1/static/app1/scss/style.scss,
            'Your/Path/project/apps/app2/static/app2/scss/style.scss
            ...
        ]
        """
        results = []
        for finder in get_finders():
            results += [finder.find(p) for p, s in finder.list(ignore_patterns) if p.endswith('.scss')]

        return results

    @staticmethod
    def get_static_paths():
        """
        Get all static paths
        :return:
        [
            'Your/Path/project/apps/app1/static',
            'Your/Path/project/apps/app2/static'
            ...
        ]
        """
        results = []
        for f in get_finders():
            if hasattr(f, 'storages'):
                results += [f.storages.get(x).location for x in f.storages if hasattr(f.storages.get(x), 'location')]

        return results

    @staticmethod
    def store_file(filename, raw):
        """
        Store sass.compile output in file
        :param filename:    'Your/Path/project/apps/users/static/users/build/css/style.css'
        :param raw:         'body {background-color: #f4f7f9;}...'
        """
        file_path = os.path.dirname(filename)
        if not os.path.exists(file_path):
            os.makedirs(file_path, exist_ok=True)

        file = open(filename, 'w', encoding='utf8')
        file.write(raw)
        file.close()

    @staticmethod
    def clean_files(filename):
        """
        Remove 'style.css', 'style.css.map', 'style.min.css', 'style.min.css.map' files from css folder
        :param filename:    'Your/Path/project/apps/users/static/users/build/css/style.css'
        """
        file, ext = os.path.splitext(filename)
        file = file[:-4] if file.endswith('.min') else file
        for extension in ('css', 'css.map', 'min.css', 'min.css.map'):
            path = '.'.join([file, extension])
            if os.path.isfile(path):
                os.remove(path)
